package cc.fascinated.backend.service;

import cc.fascinated.backend.Main;
import cc.fascinated.backend.common.DateUtils;
import cc.fascinated.backend.common.Timer;
import cc.fascinated.backend.common.WebRequest;
import cc.fascinated.backend.exception.impl.BadRequestException;
import cc.fascinated.backend.exception.impl.RateLimitException;
import cc.fascinated.backend.exception.impl.ResourceNotFoundException;
import cc.fascinated.backend.model.account.Account;
import cc.fascinated.backend.model.leaderboard.Leaderboard;
import cc.fascinated.backend.model.score.Score;
import cc.fascinated.backend.model.token.*;
import cc.fascinated.backend.repository.AccountRepository;
import cc.fascinated.backend.repository.LeaderboardRepository;
import cc.fascinated.backend.repository.ScoreRepository;
import com.google.gson.JsonObject;
import lombok.NonNull;
import lombok.SneakyThrows;
import lombok.extern.log4j.Log4j2;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.web.socket.CloseStatus;
import org.springframework.web.socket.TextMessage;
import org.springframework.web.socket.WebSocketSession;
import org.springframework.web.socket.client.standard.StandardWebSocketClient;
import org.springframework.web.socket.handler.TextWebSocketHandler;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.TimeUnit;

@Service @Log4j2(topic = "ScoreSaber Service")
public class ScoreSaberService extends TextWebSocketHandler {
    private static final long LEADERBOARD_UPDATE_INTERVAL = TimeUnit.HOURS.toMillis(24);
    private static final String SCORESABER_API = "https://scoresaber.com/api/";
    private static final String GET_PLAYER_ENDPOINT = SCORESABER_API + "player/%s/full";
    private static final String GET_PLAYER_SCORES_ENDPOINT = SCORESABER_API + "player/%s/scores?limit=100&sort=%s&page=%s&withMetadata=true";
    private static final String GET_LEADERBOARD_ENDPOINT = SCORESABER_API + "leaderboard/by-id/%s/info";

    /**
     * The {@link ScoreRepository} instance.
     */
    private final ScoreRepository scoreRepository;

    /**
     * The {@link LeaderboardRepository} instance.
     */
    private final LeaderboardRepository leaderboardRepository;

    /**
     * The {@link AccountRepository} instance.
     */
    private final AccountRepository accountRepository;

    @SneakyThrows @Autowired
    public ScoreSaberService(ScoreRepository scoreRepository, LeaderboardRepository leaderboardRepository, AccountRepository accountRepository) {
        this.scoreRepository = scoreRepository;
        this.leaderboardRepository = leaderboardRepository;
        this.accountRepository = accountRepository;

        connectWebSocket(); // Connect to the ScoreSaber WebSocket.

        Timer.scheduleRepeating(this::updateLeaderboards, LEADERBOARD_UPDATE_INTERVAL, LEADERBOARD_UPDATE_INTERVAL);
    }

    /**
     * Gets the account from the ScoreSaber API.
     *
     * @param id The id of the account.
     * @return The account.
     * @throws ResourceNotFoundException If the account is not found.
     * @throws RateLimitException If the ScoreSaber rate limit is reached.
     */
    public ScoreSaberAccountToken getAccount(String id) {
        ScoreSaberAccountToken account = WebRequest.getAsEntity(String.format(GET_PLAYER_ENDPOINT, id), ScoreSaberAccountToken.class);
        if (account == null) { // Check if the account doesn't exist.
            log.info("Account with id '{}' not found.", id);
            throw new ResourceNotFoundException("Account with id '%s' not found.".formatted(id));
        }
        if (account.isBanned()) {
            throw new BadRequestException("Account with id '%s' is banned.".formatted(id));
        }
        if (account.isInactive()) {
            throw new BadRequestException("Account with id '%s' is inactive.".formatted(id));
        }
        return account;
    }

    /**
     * Gets the scores for the account.
     *
     * @param account The account.
     * @param page The page to get the scores from.
     * @return The scores.
     */
    public ScoreSaberScoresPageToken getPageScores(Account account, int page) {
        log.info("Fetching scores for account '{}' from page {}.", account.getName(), page);
        ScoreSaberScoresPageToken pageToken = WebRequest.getAsEntity(String.format(GET_PLAYER_SCORES_ENDPOINT, account.getId(), "recent", page), ScoreSaberScoresPageToken.class);
        if (pageToken == null) { // Check if the page doesn't exist.
            return null;
        }
        // Sort the scores by newest time set.
        pageToken.setPlayerScores(Arrays.stream(pageToken.getPlayerScores())
                .sorted((a, b) -> DateUtils.getDateFromString(b.getScore().getTimeSet()).compareTo(DateUtils.getDateFromString(a.getScore().getTimeSet())))
                .toArray(ScoreSaberPlayerScoreToken[]::new));
        return pageToken;
    }

    /**
     * Gets the scores for the account.
     *
     * @param account The account.
     * @return The scores.
     */
    public List<ScoreSaberScoresPageToken> getScores(Account account) {
        List<ScoreSaberScoresPageToken> scores = new ArrayList<>(List.of(getPageScores(account, 1)));
        ScoreSaberPageMetadataToken metadata = scores.get(0).getMetadata();
        int totalPages = (int) Math.ceil((double) metadata.getTotal() / metadata.getItemsPerPage());
        log.info("Fetching {} pages of scores for account '{}'.", totalPages, account.getId());
        for (int i = 2; i <= totalPages; i++) {
            scores.add(getPageScores(account, i));
        }

        return scores;
    }

    /**
     * Fetches the scores for the account.
     *
     * @param account The account.
     */
    public void updateScores(Account account) {
        String name = account.getName();

        // Fetch the scores for the account.
        List<Score> scores = scoreRepository.getScores(account.getId());
        if (scores.isEmpty()) {
            log.warn("Account '{}' has no scores, fetching them.", name);

            List<ScoreSaberScoresPageToken> scoresPageTokens = this.getScores(account);
            List<Score> newScores = new ArrayList<>();
            List<Leaderboard> leaderboardToSave = new ArrayList<>();

            for (ScoreSaberScoresPageToken page : scoresPageTokens) {
                for (ScoreSaberPlayerScoreToken scoreToken : page.getPlayerScores()) {
                    Score score = Score.fromToken(account.getId(), scoreToken);
                    newScores.add(score);
                    scores.add(score);
                    leaderboardToSave.add(Leaderboard.fromToken(scoreToken.getLeaderboard()));
                }
            }

            // Save the leaderboards if they are missing.
            for (Leaderboard leaderboard : leaderboardToSave) {
                if (leaderboardRepository.findById(leaderboard.getId()).isEmpty()) {
                    leaderboardRepository.save(leaderboard);
                }
            }

            scoreRepository.saveAll(newScores); // Save the player's scores.
            log.info("Found {} scores for account '{}'.", newScores.size(), name);
            return;
        }

        long start = System.currentTimeMillis();
        log.info("Fetching new scores for account '{}'.", name);

        int page = 1; // The current page to search for scores.
        boolean done = false; // Whether we are done fetching scores.
        List<ScoreSaberPlayerScoreToken> newScores = new ArrayList<>();
        do {
            // This will keep fetching score pages until it finds a score that already exists.
            ScoreSaberScoresPageToken pageScores = getPageScores(account, page);
            for (ScoreSaberPlayerScoreToken score : pageScores.getPlayerScores()) {
                boolean exists = scores.stream().anyMatch(s -> s.getId().equals(score.getScore().getId()));
                if (!exists) {
                    newScores.add(score);
                    continue;
                }
                done = true;
            }
            page++;
        } while (!done);

        // Save the new scores.
        for (ScoreSaberPlayerScoreToken score : newScores) {
            saveScore(account, score);
        }

       log.info("Found {} new scores for account '{}'. (took: {}ms)", newScores.size(), name, System.currentTimeMillis() - start);
    }

    /**
     * Saves the score for the account.
     *
     * @param account The account.
     * @param score   The score to save.
     */
    private void saveScore(Account account, ScoreSaberPlayerScoreToken score) {
        Leaderboard newScoreLeaderboard = Leaderboard.fromToken(score.getLeaderboard());
        Score oldScore = scoreRepository.findById(score.getScore().getId()).orElse(null);

        // The score has an old score.
        if (oldScore != null) {
            Leaderboard oldScoreLeaderboard = leaderboardRepository.findById(oldScore.getLeaderboardId()).orElse(null);

            if (oldScoreLeaderboard != null && oldScoreLeaderboard.getId().equals(newScoreLeaderboard.getId())) {
                // If it matches, add the new score and retain information about the old score.
                Score scoreSet = Score.fromToken(account.getId(), score);
                oldScore.setPreviousScores(null); // We don't want nested previous scores.
                scoreSet.getPreviousScores().add(oldScore);

                scoreRepository.delete(oldScore); // Delete the old score.
                scoreRepository.save(scoreSet); // Save the new score.
            }
        } else {
            // The score is new
            scoreRepository.save(Score.fromToken(account.getId(), score)); // Save the new score.
        }

        // Check if the leaderboard doesn't already exist.
        if (leaderboardRepository.findById(newScoreLeaderboard.getId()).isEmpty()) {
            leaderboardRepository.save(newScoreLeaderboard); // Save the leaderboard.
        }
    }

    /**
     * Updates the leaderboards.
     */
    private void updateLeaderboards() {
        log.info("Updating leaderboards.");
        List<String> leaderboardIds = new ArrayList<>();

        // Get all the unique leaderboard ids.
        for (Score score : scoreRepository.findAll()) {
            leaderboardIds.add(score.getLeaderboardId());
        }

        // Fetch the leaderboards.
        for (String leaderboardId : leaderboardIds) {
            ScoreSaberLeaderboardToken leaderboard = WebRequest.getAsEntity(String.format(GET_LEADERBOARD_ENDPOINT, leaderboardId),
                    ScoreSaberLeaderboardToken.class);

            // No leaderboard found.
            if (leaderboard == null) {
                log.warn("Leaderboard '{}' not found.", leaderboardId);
                continue;
            }

            // Save the leaderboard.
            leaderboardRepository.save(Leaderboard.fromToken(leaderboard));
        }
        log.info("Updated {} leaderboards.", leaderboardIds.size());
    }

    /**
     * Connects to the ScoreSaber WebSocket.
     */
    @SneakyThrows
    private void connectWebSocket() {
        log.info("Connecting to the ScoreSaber WSS.");
        new StandardWebSocketClient().execute(this, "wss://scoresaber.com/ws").get();
    }

    @Override
    public void afterConnectionEstablished(@NonNull WebSocketSession session) {
        log.info("Connected to the ScoreSaber WSS.");
    }

    @Override
    public void afterConnectionClosed(@NonNull WebSocketSession session, @NonNull CloseStatus status) {
        log.info("Disconnected from the ScoreSaber WSS.");
        connectWebSocket(); // Reconnect to the WebSocket.
    }

    @Override @SneakyThrows
    protected void handleTextMessage(@NonNull WebSocketSession session, @NonNull TextMessage message) {
        // Ignore the connection message.
        if (message.getPayload().equals("Connected to the ScoreSaber WSS")) {
            return;
        }

        try {
            JsonObject json = Main.GSON.fromJson(message.getPayload(), JsonObject.class);
            String command = json.get("commandName").getAsString();
            JsonObject data = json.get("commandData").getAsJsonObject();

            if (command.equals("score")) {
                ScoreSaberPlayerScoreToken score = Main.GSON.fromJson(data, ScoreSaberPlayerScoreToken.class);
                ScoreSaberScoreToken.LeaderboardPlayerInfo playerInfo = score.getScore().getLeaderboardPlayerInfo();

                // Fetch the account.
                Optional<Account> account = accountRepository.findById(playerInfo.getId());
                if (account.isEmpty()) {
                    // We don't track this account, so ignore it.
                    return;
                }

                // Save the score.
                saveScore(account.get(), score);
                log.info("Saved websocket score '{}' for account '{}'.", score.getScore().getId(), playerInfo.getName());
            }
        } catch (Exception ex) {
            log.error("An error occurred while handling the message.", ex);
        }
    }
}
